/* This file is part of Vcsync
   Copyright (C) 2014 Sergey Poznyakoff
  
   Vcsync is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   Vcsync is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with Vcsync.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "vcsync.h"
#include <signal.h>

int dry_run;
int debug_level;
int lint_only;
int stderr_only;
char *config_file;
char *destination_root;
unsigned sleep_option;
int detach_option;
char *vcs_type;

struct vcs_tab {
	const char *name;
	vcs_proc_fn proc;
};

static struct vcs_tab vcs_tab[] = {
	{ "cvs", cvs_proc },
	{ "svn", svn_proc },
	{ "git", git_proc },
	{ NULL }
};

vcs_proc_fn
vcs_proc_get(const char *name)
{
	struct vcs_tab *p;

	for (p = vcs_tab; p->name; p++)
		if (strcmp(p->name, name) == 0)
			break;
	
	return p->proc;
}

char *
pathcat(const char **path)
{
	size_t len = 0;
	const char **pp;
	char *ret, *p;
	const char *q;
	
	for (pp = path; *pp; pp++)
		len += strlen(*pp) + !!pp[1];

	ret = emalloc(len + 1);
	for (pp = path, p = ret; *pp; pp++) {
		for (q = *pp; *p = *q; p++, q++)
			;
		if (pp[1])
			*p++ = '/';
	}
	*p = 0;
	
	return ret;
}
	
char *
mkfilename(const char *dir, const char *name)
{
	const char *a[3];
	
	a[0] = dir;
	a[1] = name;
	a[2] = NULL;
	return pathcat(a);
}


#include "cmdline.h"

int
main(int argc, char **argv)
{
	int i;
	struct grecs_node *tree = NULL;

	setprogname(argv[0]);
	config_init();
	parse_options(argc, argv, &i);

	argc -= i;
	argv += i;

	if (!config_file) {
		if (access(DEFAULT_CONFIG_FILE, F_OK) == 0)
			config_file = DEFAULT_CONFIG_FILE;
		else {
			if (lint_only)
				die(EX_USAGE,
				    "configuration file %s doesn't exist",
				    DEFAULT_CONFIG_FILE);
			debug(1, ("default configuration file %s does not "
				  "exist", DEFAULT_CONFIG_FILE));
		}
	}

	if (config_file) {
		tree = grecs_parse(config_file);
		if (!tree)
			exit(EX_CONFIG);
	}
	
	config_finish(tree);
	if (lint_only)
		return EX_OK;

	if (have_checkoutrc()) {
		diag(LOG_INFO, "testing checkoutrc files");
		flush_checkoutrc_files(NULL);
		exit(EX_OK);
	}

	/* Fixup options */
	if (config->detach)
		stderr_only = 0;
	
        /* Reset log_to_stderr to whatever the user wants (if -l was
	   given), or to the default -1.
        */
	log_to_stderr = cli_log_to_stderr;
	
	if (!stderr_only && cfg_facility > 0) {
		openlog(cfg_tag, LOG_PID, cfg_facility);
		grecs_log_to_stderr = 0;
	}

	if (config->message && *config->message) {
		fputs(config->message, stderr);
		if (config->message[strlen(config->message)-1] != '\n')
			fputc('\n', stderr);
	}
	
	if (config->detach) {
		if (daemon(1, 1))
			die(EX_FAIL, "daemon: %s", strerror(errno));
		log_to_stderr = 0;
	}

	if (config->sleep_time)
		sleep(config->sleep_time);

	sigsetup(sigmain);
	flush_checkoutrc_files(config->proc(argc, argv));
	return EX_OK;
}
